<?php

namespace Drupal\pbs_media_manager\Commands;

use Drupal\pbs_media_manager\EpisodeManager;
use Drupal\pbs_media_manager\SeasonManager;
use Drupal\pbs_media_manager\ShowManager;
use Drush\Commands\DrushCommands;

/**
 * A Drush commandfile.
 */
class PbsCommands extends DrushCommands {

  /**
   * The PBS Media Manager show manager.
   *
   * @var \Drupal\pbs_media_manager\ShowManager
   */
  protected $showManager;

  /**
   * The PBS Media Manager season manager.
   *
   * @var \Drupal\pbs_media_manager\SeasonManager
   */
  protected $seasonManager;

  /**
   * The PBS Media Manager episode manager.
   *
   * @var \Drupal\pbs_media_manager\EpisodeManager
   */
  protected $episodeManager;

  /**
   * Constructs a new DrushCommands object.
   *
   * @param \Drupal\pbs_media_manager\ShowManager $show_manager
   *   The PBS Media Manager show manager.
   * @param \Drupal\pbs_media_manager\SeasonManager $season_manager
   *   The PBS Media Manager season manager.
   * @param \Drupal\pbs_media_manager\EpisodeManager $episode_manager
   *   The PBS Media Manager episode manager.
   */
  public function __construct(ShowManager $show_manager, SeasonManager $season_manager, EpisodeManager $episode_manager) {
    $this->showManager = $show_manager;
    $this->seasonManager = $season_manager;
    $this->episodeManager = $episode_manager;
  }

  /**
   * Command to pull a single Media Manager show.
   *
   * @param string $show_id
   *   The Media Manager id of the show.
   * @param array $options
   *   Associative array of options.
   *
   * @command pbs_media_manager:getShow
   * @aliases pbs-gs
   */
  public function getShow($show_id, array $options = ['force' => FALSE]) {
    $show = $this->showManager->getShow($show_id);
    if (is_object($show)) {
      $this->showManager->addOrUpdateShow($show, $options['force']);
    }
    else {
      $this->logger->error(dt('The show ID @id did not return any results.', [
        '@id' => $show_id,
      ]));
    }
  }

  /**
   * Command to pull a single Media Manager season.
   *
   * @param string $season_id
   *   The Media Manager id of the season.
   * @param array $options
   *   Associative array of options.
   *
   * @command pbs_media_manager:getSeason
   * @aliases pbs-gn
   */
  public function getSeason($season_id, array $options = ['force' => FALSE]) {
    $season = $this->seasonManager->getSeason($season_id);
    if (is_object($season)) {
      $this->seasonManager->addOrUpdateSeason($season, $options['force']);
    }
    else {
      $this->logger->error(dt('The season ID @id did not return any results.', [
        '@id' => $season_id,
      ]));
    }
  }

  /**
   * Command to pull a single Media Manager episode.
   *
   * @param string $episode_id
   *   The Media Manager id of the episode.
   * @param array $options
   *   Associative array of options.
   *
   * @usage drush pbs-gv 745a2571-05fd-449e-93b6-0b8fb89e6fa1 --force
   *   Import the episode or force it to update even if it has not changed.
   *
   * @command pbs_media_manager:getEpisode
   * @aliases pbs-ge
   */
  public function getEpisode($episode_id, array $options = ['force' => FALSE]) {
    $episode = $this->episodeManager->getEpisode($episode_id);
    if (is_object($episode)) {
      $this->episodeManager->addOrUpdateEpisode($episode, $options['force']);
    }
    else {
      $this->logger->error(dt('The episode ID @id did not return any results.', [
        '@id' => $episode_id,
      ]));
    }
  }

  /**
   * Command to pull a single Media Manager video.
   *
   * @param string $video_id
   *   The Media Manager id of the video.
   * @param array $options
   *   Associative array of options.
   *
   * @usage drush pbs-gv 82bb45da-12de-4a52-8ff9-5ee02dd30442 --force
   *   Import the video or force it to update even if it has not changed.
   *
   * @command pbs_media_manager:getVideo
   * @aliases pbs-gv
   */
  public function getVideo($video_id, array $options = ['force' => FALSE]) {
    // This sort of hijacks the EpisodeManager class.
    $client = $this->episodeManager->getApiClient();
    $video = $client->getAsset($video_id);
    if (is_object($video)) {
      $this->episodeManager->addOrUpdateVideoContent($video, $options['force']);
    }
    else {
      $this->logger->error(dt('The video ID @id did not return any results.', [
        '@id' => $video_id,
      ]));
    }
  }

}
