<?php

namespace Drupal\pbs_media_manager\Form;

use Drupal\Core\CronInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Queue\QueueFactory;
use Drupal\node\NodeInterface;
use Drupal\pbs_media_manager\EpisodeManager;
use Drupal\pbs_media_manager\SeasonManager;
use Drupal\pbs_media_manager\ShowManager;
use Symfony\Component\DependencyInjection\ContainerInterface;

class ShowUpdateBlockForm extends FormBase {

  /**
   * The PBS Media Manager show manager.
   *
   * @var \Drupal\pbs_media_manager\ShowManager
   */
  protected $showManager;

  /**
   * The PBS Media Manager season manager.
   *
   * @var \Drupal\pbs_media_manager\SeasonManager
   */
  protected $seasonManager;

  /**
   * The PBS Media Manager episode manager.
   *
   * @var \Drupal\pbs_media_manager\EpisodeManager
   */
  protected $episodeManager;

  /**
   * The messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * Queue service.
   *
   * @var \Drupal\Core\Queue\QueueFactory
   */
  protected $queueFactory;

  /**
   * The cron service.
   *
   * @var \Drupal\Core\CronInterface
   */
  protected $cron;

  /**
   * Constructs a new DrushCommands object.
   *
   * @param \Drupal\pbs_media_manager\ShowManager $show_manager
   *   The PBS Media Manager show manager.
   * @param \Drupal\pbs_media_manager\SeasonManager $season_manager
   *   The PBS Media Manager season manager.
   * @param \Drupal\pbs_media_manager\EpisodeManager $episode_manager
   *   The PBS Media Manager episode manager.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger service.
   * @param \Drupal\Core\Queue\QueueFactory $queue_factory
   *   Queue factory service.
   * @param \Drupal\Core\CronInterface $cron
   *   The cron service.
   */
  public function __construct(ShowManager $show_manager, SeasonManager $season_manager, EpisodeManager $episode_manager, MessengerInterface $messenger, QueueFactory $queue_factory, CronInterface $cron) {
    $this->showManager = $show_manager;
    $this->seasonManager = $season_manager;
    $this->episodeManager = $episode_manager;
    $this->messenger = $messenger;
    $this->queueFactory = $queue_factory;
    $this->cron = $cron;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('pbs_media_manager.show_manager'),
      $container->get('pbs_media_manager.season_manager'),
      $container->get('pbs_media_manager.episode_manager'),
      $container->get('messenger'),
      $container->get('queue'),
      $container->get('cron')
    );
  }

  /**
   * {@inheritdoc}.
   */
  public function getFormId() {
    return 'show_update_block_form';
  }

  /**
   * {@inheritdoc}.
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $form['update'] = [
      '#type' => 'submit',
      '#value' => $this->t('Update Show'),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $node = $this->getRouteMatch()->getParameter('node');

    $config = $this->config('pbs_media_manager.settings');
    $show_id_field = $config->get('shows.mappings.id');
    if (empty($show_id_field)) {
      $this->messenger->addWarning($this->t('The "id" value in the "Show field
        mappings" must be configured to update this show.'));
      return;
    }

    if ($node instanceof NodeInterface) {
      $uuid = $node->get($show_id_field)->value;
      $title = $node->getTitle();
    }

    if (empty($uuid)) {
      $this->messenger->addWarning($this->t('The Media Manager ID field is empty for the show "@title".', ['@title' => $title]));
      return;
    }

    // First, update the show.
    $show = $this->showManager->getShow($uuid);
    if (!is_object($show)) {
      $this->messenger->addWarning($this->t('The show "@title" was not found in Media Manager.', ['@title' => $title]));
      return;
    }
    $showQueue = $this->queueFactory->get('pbs_media_manager.queue.shows');
    $showQueue->createItem($show);
    $this->messenger->addStatus(
      $this->t('The show "@title" has been added to the queue.', [
        '@title' => $title,
      ]
    ));
    // Second, if the show has seasons, update all of them.
    if ($seasons = $show->attributes->seasons) {
      foreach ($seasons as $season_item) {
        $season = $this->seasonManager->getSeason($season_item->id);
        if (is_object($season)) {
          $seasonQueue = $this->queueFactory->get('pbs_media_manager.queue.seasons');
          $seasonQueue->createItem($season);
          $this->messenger->addStatus(
            $this->t('Season @ordinal has been added to the queue.', [
              '@ordinal' => $season->attributes->ordinal,
            ]
          ));
        }

        // Third, update all of the episodes for each season.
        if ($episodes = $season->attributes->episodes) {
          foreach ($episodes as $episode_item) {
            $episode = $this->episodeManager->getEpisode($episode_item->id);
            if (is_object($episode)) {
              $episodeQueue = $this->queueFactory->get('pbs_media_manager.queue.episodes');
              $episodeQueue->createItem($episode);
              $this->messenger->addStatus(
                $this->t('Episode "@title" has been added to the queue.', [
                  '@title' => $episode->attributes->title,
                ]
              ));
            }
          }
        }
      }
    }
    if ($this->cron->run()) {
      $this->messenger()->addStatus($this->t('Cron ran successfully.'));
    }
    else {
      $this->messenger()->addError($this->t('Cron run failed.'));
    }
  }

}
